@@ -18,18 +18,22 @@
 // FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 // DEALINGS IN THE SOFTWARE.
 
-use futures_rustls::{webpki, client, server};
 use crate::{error::Error, tls};
 use either::Either;
 use futures::{future::BoxFuture, prelude::*, ready, stream::BoxStream};
+use futures_rustls::{client, rustls, server};
 use libp2p_core::{
-    Transport,
     either::EitherOutput,
-    multiaddr::{Protocol, Multiaddr},
-    transport::{ListenerEvent, TransportError}
+    multiaddr::{Multiaddr, Protocol},
+    transport::{ListenerEvent, TransportError},
+    Transport,
 };
 use log::{debug, trace};
-use soketto::{connection, extension::deflate::Deflate, handshake};
+use soketto::{
+    connection::{self, CloseReason},
+    extension::deflate::Deflate,
+    handshake,
+};
 use std::{convert::TryInto, fmt, io, mem, pin::Pin, task::Context, task::Poll};
 use url::Url;
 
@@ -45,7 +49,7 @@ pub struct WsConfig<T> {
     max_data_size: usize,
     tls_config: tls::Config,
     max_redirects: u8,
-    use_deflate: bool
+    use_deflate: bool,
 }
 
 impl<T> WsConfig<T> {
@@ -56,7 +60,7 @@ impl<T> WsConfig<T> {
             max_data_size: MAX_DATA_SIZE,
             tls_config: tls::Config::client(),
             max_redirects: 0,
-            use_deflate: false
+            use_deflate: false,
         }
     }
 
@@ -104,11 +108,12 @@ where
     T::Dial: Send + 'static,
     T::Listener: Send + 'static,
     T::ListenerUpgrade: Send + 'static,
-    T::Output: AsyncRead + AsyncWrite + Unpin + Send + 'static
+    T::Output: AsyncRead + AsyncWrite + Unpin + Send + 'static,
 {
     type Output = Connection<T::Output>;
     type Error = Error<T::Error>;
-    type Listener = BoxStream<'static, Result<ListenerEvent<Self::ListenerUpgrade, Self::Error>, Self::Error>>;
+    type Listener =
+        BoxStream<'static, Result<ListenerEvent<Self::ListenerUpgrade, Self::Error>, Self::Error>>;
     type ListenerUpgrade = BoxFuture<'static, Result<Self::Output, Self::Error>>;
     type Dial = BoxFuture<'static, Result<Self::Output, Self::Error>>;
 
@@ -116,24 +121,28 @@ where
         let mut inner_addr = addr.clone();
 
         let (use_tls, proto) = match inner_addr.pop() {
-            Some(p@Protocol::Wss(_)) =>
+            Some(p @ Protocol::Wss(_)) => {
                 if self.tls_config.server.is_some() {
                     (true, p)
                 } else {
                     debug!("/wss address but TLS server support is not configured");
-                    return Err(TransportError::MultiaddrNotSupported(addr))
+                    return Err(TransportError::MultiaddrNotSupported(addr));
                 }
-            Some(p@Protocol::Ws(_)) => (false, p),
+            }
+            Some(p @ Protocol::Ws(_)) => (false, p),
             _ => {
                 debug!("{} is not a websocket multiaddr", addr);
-                return Err(TransportError::MultiaddrNotSupported(addr))
+                return Err(TransportError::MultiaddrNotSupported(addr));
             }
         };
 
         let tls_config = self.tls_config;
         let max_size = self.max_data_size;
         let use_deflate = self.use_deflate;
-        let transport = self.transport.listen_on(inner_addr).map_err(|e| e.map(Error::Transport))?;
+        let transport = self
+            .transport
+            .listen_on(inner_addr)
+            .map_err(|e| e.map(Error::Transport))?;
         let listen = transport
             .map_err(Error::Transport)
             .map_ok(move |event| match event {
@@ -146,10 +155,12 @@ where
                     a = a.with(proto.clone());
                     ListenerEvent::AddressExpired(a)
                 }
-                ListenerEvent::Error(err) => {
-                    ListenerEvent::Error(Error::Transport(err))
-                }
-                ListenerEvent::Upgrade { upgrade, mut local_addr, mut remote_addr } => {
+                ListenerEvent::Error(err) => ListenerEvent::Error(Error::Transport(err)),
+                ListenerEvent::Upgrade {
+                    upgrade,
+                    mut local_addr,
+                    mut remote_addr,
+                } => {
                     local_addr = local_addr.with(proto.clone());
                     remote_addr = remote_addr.with(proto.clone());
                     let remote1 = remote_addr.clone(); // used for logging
@@ -160,28 +171,30 @@ where
                         let stream = upgrade.map_err(Error::Transport).await?;
                         trace!("incoming connection from {}", remote1);
 
-                        let stream =
-                            if use_tls { // begin TLS session
-                                let server = tls_config
-                                    .server
-                                    .expect("for use_tls we checked server is not none");
+                        let stream = if use_tls {
+                            // begin TLS session
+                            let server = tls_config
+                                .server
+                                .expect("for use_tls we checked server is not none");
 
-                                trace!("awaiting TLS handshake with {}", remote1);
+                            trace!("awaiting TLS handshake with {}", remote1);
 
-                                let stream = server.accept(stream)
-                                    .map_err(move |e| {
-                                        debug!("TLS handshake with {} failed: {}", remote1, e);
-                                        Error::Tls(tls::Error::from(e))
-                                    })
-                                    .await?;
+                            let stream = server
+                                .accept(stream)
+                                .map_err(move |e| {
+                                    debug!("TLS handshake with {} failed: {}", remote1, e);
+                                    Error::Tls(tls::Error::from(e))
+                                })
+                                .await?;
 
-                                let stream: TlsOrPlain<_> =
-                                    EitherOutput::First(EitherOutput::Second(stream));
+                            let stream: TlsOrPlain<_> =
+                                EitherOutput::First(EitherOutput::Second(stream));
 
-                                stream
-                            } else { // continue with plain stream
-                                EitherOutput::Second(stream)
-                            };
+                            stream
+                        } else {
+                            // continue with plain stream
+                            EitherOutput::Second(stream)
+                        };
 
                         trace!("receiving websocket handshake request from {}", remote2);
 
@@ -192,21 +205,22 @@ where
                         }
 
                         let ws_key = {
-                            let request = server.receive_request()
+                            let request = server
+                                .receive_request()
                                 .map_err(|e| Error::Handshake(Box::new(e)))
                                 .await?;
-                            request.into_key()
+                            request.key()
                         };
 
                         trace!("accepting websocket handshake request from {}", remote2);
 
-                        let response =
-                            handshake::server::Response::Accept {
-                                key: &ws_key,
-                                protocol: None
-                            };
+                        let response = handshake::server::Response::Accept {
+                            key: ws_key,
+                            protocol: None,
+                        };
 
-                        server.send_response(&response)
+                        server
+                            .send_response(&response)
                             .map_err(|e| Error::Handshake(Box::new(e)))
                             .await?;
 
@@ -223,7 +237,7 @@ where
                     ListenerEvent::Upgrade {
                         upgrade: Box::pin(upgrade) as BoxFuture<'static, _>,
                         local_addr,
-                        remote_addr
+                        remote_addr,
                     }
                 }
             });
@@ -233,7 +247,9 @@ where
     fn dial(self, addr: Multiaddr) -> Result<Self::Dial, TransportError<Self::Error>> {
         let addr = match parse_ws_dial_addr(addr) {
             Ok(addr) => addr,
-            Err(Error::InvalidMultiaddr(a)) => return Err(TransportError::MultiaddrNotSupported(a)),
+            Err(Error::InvalidMultiaddr(a)) => {
+                return Err(TransportError::MultiaddrNotSupported(a))
+            }
             Err(e) => return Err(TransportError::Other(e)),
         };
 
@@ -247,13 +263,13 @@ where
                     Ok(Either::Left(redirect)) => {
                         if remaining_redirects == 0 {
                             debug!("Too many redirects (> {})", self.max_redirects);
-                            return Err(Error::TooManyRedirects)
+                            return Err(Error::TooManyRedirects);
                         }
                         remaining_redirects -= 1;
                         addr = parse_ws_dial_addr(location_to_multiaddr(&redirect)?)?
                     }
                     Ok(Either::Right(conn)) => return Ok(conn),
-                    Err(e) => return Err(e)
+                    Err(e) => return Err(e),
                 }
             }
         };
@@ -269,37 +285,45 @@ where
 impl<T> WsConfig<T>
 where
     T: Transport,
-    T::Output: AsyncRead + AsyncWrite + Send + Unpin + 'static
+    T::Output: AsyncRead + AsyncWrite + Send + Unpin + 'static,
 {
     /// Attempts to dial the given address and perform a websocket handshake.
-    async fn dial_once(self, addr: WsAddress) -> Result<Either<String, Connection<T::Output>>, Error<T::Error>> {
+    async fn dial_once(
+        self,
+        addr: WsAddress,
+    ) -> Result<Either<String, Connection<T::Output>>, Error<T::Error>> {
         trace!("Dialing websocket address: {:?}", addr);
 
-        let dial = self.transport.dial(addr.tcp_addr)
-            .map_err(|e| match e {
-                TransportError::MultiaddrNotSupported(a) => Error::InvalidMultiaddr(a),
-                TransportError::Other(e) => Error::Transport(e)
-            })?;
+        let dial = self.transport.dial(addr.tcp_addr).map_err(|e| match e {
+            TransportError::MultiaddrNotSupported(a) => Error::InvalidMultiaddr(a),
+            TransportError::Other(e) => Error::Transport(e),
+        })?;
 
         let stream = dial.map_err(Error::Transport).await?;
         trace!("TCP connection to {} established.", addr.host_port);
 
-        let stream =
-            if addr.use_tls { // begin TLS session
-                let dns_name = addr.dns_name.expect("for use_tls we have checked that dns_name is some");
-                trace!("Starting TLS handshake with {:?}", dns_name);
-                let stream = self.tls_config.client.connect(dns_name.as_ref(), stream)
-                    .map_err(|e| {
-                        debug!("TLS handshake with {:?} failed: {}", dns_name, e);
-                        Error::Tls(tls::Error::from(e))
-                    })
-                    .await?;
+        let stream = if addr.use_tls {
+            // begin TLS session
+            let dns_name = addr
+                .dns_name
+                .expect("for use_tls we have checked that dns_name is some");
+            trace!("Starting TLS handshake with {:?}", dns_name);
+            let stream = self
+                .tls_config
+                .client
+                .connect(dns_name.clone(), stream)
+                .map_err(|e| {
+                    debug!("TLS handshake with {:?} failed: {}", dns_name, e);
+                    Error::Tls(tls::Error::from(e))
+                })
+                .await?;
 
-                let stream: TlsOrPlain<_> = EitherOutput::First(EitherOutput::First(stream));
-                stream
-            } else { // continue with plain stream
-                EitherOutput::Second(stream)
-            };
+            let stream: TlsOrPlain<_> = EitherOutput::First(EitherOutput::First(stream));
+            stream
+        } else {
+            // continue with plain stream
+            EitherOutput::Second(stream)
+        };
 
         trace!("Sending websocket handshake to {}", addr.host_port);
 
@@ -309,9 +333,19 @@ where
             client.add_extension(Box::new(Deflate::new(connection::Mode::Client)));
         }
 
-        match client.handshake().map_err(|e| Error::Handshake(Box::new(e))).await? {
-            handshake::ServerResponse::Redirect { status_code, location } => {
-                debug!("received redirect ({}); location: {}", status_code, location);
+        match client
+            .handshake()
+            .map_err(|e| Error::Handshake(Box::new(e)))
+            .await?
+        {
+            handshake::ServerResponse::Redirect {
+                status_code,
+                location,
+            } => {
+                debug!(
+                    "received redirect ({}); location: {}",
+                    status_code, location
+                );
                 Ok(Either::Left(location))
             }
             handshake::ServerResponse::Rejected { status_code } => {
@@ -330,7 +364,7 @@ where
 struct WsAddress {
     host_port: String,
     path: String,
-    dns_name: Option<webpki::DNSName>,
+    dns_name: Option<rustls::ServerName>,
     use_tls: bool,
     tcp_addr: Multiaddr,
 }
@@ -349,20 +383,23 @@ fn parse_ws_dial_addr<T>(addr: Multiaddr) -> Result<WsAddress, Error<T>> {
     let mut tcp = protocols.next();
     let (host_port, dns_name) = loop {
         match (ip, tcp) {
-            (Some(Protocol::Ip4(ip)), Some(Protocol::Tcp(port)))
-                => break (format!("{}:{}", ip, port), None),
-            (Some(Protocol::Ip6(ip)), Some(Protocol::Tcp(port)))
-                => break (format!("{}:{}", ip, port), None),
-            (Some(Protocol::Dns(h)), Some(Protocol::Tcp(port))) |
-            (Some(Protocol::Dns4(h)), Some(Protocol::Tcp(port))) |
-            (Some(Protocol::Dns6(h)), Some(Protocol::Tcp(port))) |
-            (Some(Protocol::Dnsaddr(h)), Some(Protocol::Tcp(port)))
-                => break (format!("{}:{}", &h, port), Some(tls::dns_name_ref(&h)?.to_owned())),
+            (Some(Protocol::Ip4(ip)), Some(Protocol::Tcp(port))) => {
+                break (format!("{}:{}", ip, port), None)
+            }
+            (Some(Protocol::Ip6(ip)), Some(Protocol::Tcp(port))) => {
+                break (format!("{}:{}", ip, port), None)
+            }
+            (Some(Protocol::Dns(h)), Some(Protocol::Tcp(port)))
+            | (Some(Protocol::Dns4(h)), Some(Protocol::Tcp(port)))
+            | (Some(Protocol::Dns6(h)), Some(Protocol::Tcp(port)))
+            | (Some(Protocol::Dnsaddr(h)), Some(Protocol::Tcp(port))) => {
+                break (format!("{}:{}", &h, port), Some(tls::dns_name_ref(&h)?))
+            }
             (Some(_), Some(p)) => {
                 ip = Some(p);
                 tcp = protocols.next();
             }
-            _ => return Err(Error::InvalidMultiaddr(addr))
+            _ => return Err(Error::InvalidMultiaddr(addr)),
         }
     };
 
@@ -373,16 +410,16 @@ fn parse_ws_dial_addr<T>(addr: Multiaddr) -> Result<WsAddress, Error<T>> {
     let mut p2p = None;
     let (use_tls, path) = loop {
         match protocols.pop() {
-            p@Some(Protocol::P2p(_)) => { p2p = p }
+            p @ Some(Protocol::P2p(_)) => p2p = p,
             Some(Protocol::Ws(path)) => break (false, path.into_owned()),
             Some(Protocol::Wss(path)) => {
                 if dns_name.is_none() {
                     debug!("Missing DNS name in WSS address: {}", addr);
-                    return Err(Error::InvalidMultiaddr(addr))
+                    return Err(Error::InvalidMultiaddr(addr));
                 }
-                break (true, path.into_owned())
+                break (true, path.into_owned());
             }
-            _ => return Err(Error::InvalidMultiaddr(addr))
+            _ => return Err(Error::InvalidMultiaddr(addr)),
         }
     };
 
@@ -390,7 +427,7 @@ fn parse_ws_dial_addr<T>(addr: Multiaddr) -> Result<WsAddress, Error<T>> {
     // makes up the the address for the inner TCP-based transport.
     let tcp_addr = match p2p {
         Some(p) => protocols.with(p),
-        None => protocols
+        None => protocols,
     };
 
     Ok(WsAddress {
@@ -408,16 +445,10 @@ fn location_to_multiaddr<T>(location: &str) -> Result<Multiaddr, Error<T>> {
         Ok(url) => {
             let mut a = Multiaddr::empty();
             match url.host() {
-                Some(url::Host::Domain(h)) => {
-                    a.push(Protocol::Dns(h.into()))
-                }
-                Some(url::Host::Ipv4(ip)) => {
-                    a.push(Protocol::Ip4(ip))
-                }
-                Some(url::Host::Ipv6(ip)) => {
-                    a.push(Protocol::Ip6(ip))
-                }
-                None => return Err(Error::InvalidRedirectLocation)
+                Some(url::Host::Domain(h)) => a.push(Protocol::Dns(h.into())),
+                Some(url::Host::Ipv4(ip)) => a.push(Protocol::Ip4(ip)),
+                Some(url::Host::Ipv6(ip)) => a.push(Protocol::Ip6(ip)),
+                None => return Err(Error::InvalidRedirectLocation),
             }
             if let Some(p) = url.port() {
                 a.push(Protocol::Tcp(p))
@@ -429,7 +460,7 @@ fn location_to_multiaddr<T>(location: &str) -> Result<Multiaddr, Error<T>> {
                 a.push(Protocol::Ws(url.path().into()))
             } else {
                 debug!("unsupported scheme: {}", s);
-                return Err(Error::InvalidRedirectLocation)
+                return Err(Error::InvalidRedirectLocation);
             }
             Ok(a)
         }
@@ -442,55 +473,68 @@ fn location_to_multiaddr<T>(location: &str) -> Result<Multiaddr, Error<T>> {
 
 /// The websocket connection.
 pub struct Connection<T> {
-    receiver: BoxStream<'static, Result<IncomingData, connection::Error>>,
+    receiver: BoxStream<'static, Result<Incoming, connection::Error>>,
     sender: Pin<Box<dyn Sink<OutgoingData, Error = connection::Error> + Send>>,
-    _marker: std::marker::PhantomData<T>
+    _marker: std::marker::PhantomData<T>,
 }
 
-/// Data received over the websocket connection.
+/// Data or control information received over the websocket connection.
 #[derive(Debug, Clone)]
-pub enum IncomingData {
-    /// Binary application data.
-    Binary(Vec<u8>),
-    /// UTF-8 encoded application data.
-    Text(Vec<u8>),
+pub enum Incoming {
+    /// Application data.
+    Data(Data),
     /// PONG control frame data.
-    Pong(Vec<u8>)
+    Pong(Vec<u8>),
+    /// Close reason.
+    Closed(CloseReason),
 }
 
-impl IncomingData {
+/// Application data received over the websocket connection
+#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
+pub enum Data {
+    /// UTF-8 encoded textual data.
+    Text(Vec<u8>),
+    /// Binary data.
+    Binary(Vec<u8>),
+}
+
+impl Data {
+    pub fn into_bytes(self) -> Vec<u8> {
+        match self {
+            Data::Text(d) => d,
+            Data::Binary(d) => d,
+        }
+    }
+}
+
+impl AsRef<[u8]> for Data {
+    fn as_ref(&self) -> &[u8] {
+        match self {
+            Data::Text(d) => d,
+            Data::Binary(d) => d,
+        }
+    }
+}
+
+impl Incoming {
     pub fn is_data(&self) -> bool {
         self.is_binary() || self.is_text()
     }
 
     pub fn is_binary(&self) -> bool {
-        if let IncomingData::Binary(_) = self { true } else { false }
+        matches!(self, Incoming::Data(Data::Binary(_)))
     }
 
     pub fn is_text(&self) -> bool {
-        if let IncomingData::Text(_) = self { true } else { false }
+        matches!(self, Incoming::Data(Data::Text(_)))
     }
 
     pub fn is_pong(&self) -> bool {
-        if let IncomingData::Pong(_) = self { true } else { false }
+        matches!(self, Incoming::Pong(_))
     }
 
-    pub fn into_bytes(self) -> Vec<u8> {
-        match self {
-            IncomingData::Binary(d) => d,
-            IncomingData::Text(d) => d,
-            IncomingData::Pong(d) => d
-        }
-    }
-}
-
-impl AsRef<[u8]> for IncomingData {
-    fn as_ref(&self) -> &[u8] {
-        match self {
-            IncomingData::Binary(d) => d,
-            IncomingData::Text(d) => d,
-            IncomingData::Pong(d) => d
-        }
+    pub fn is_close(&self) -> bool {
+        matches!(self, Incoming::Closed(_))
     }
 }
 
@@ -503,7 +547,7 @@ pub enum OutgoingData {
     Ping(Vec<u8>),
     /// Send an unsolicited PONG message.
     /// (Incoming PINGs are answered automatically.)
-    Pong(Vec<u8>)
+    Pong(Vec<u8>),
 }
 
 impl<T> fmt::Debug for Connection<T> {
@@ -514,7 +558,7 @@ impl<T> fmt::Debug for Connection<T> {
 
 impl<T> Connection<T>
 where
-    T: AsyncRead + AsyncWrite + Send + Unpin + 'static
+    T: AsyncRead + AsyncWrite + Send + Unpin + 'static,
 {
     fn new(builder: connection::Builder<TlsOrPlain<T>>) -> Self {
         let (sender, receiver) = builder.finish();
@@ -536,29 +580,34 @@ where
                     sender.send_pong(data).await?
                 }
                 quicksink::Action::Flush => sender.flush().await?,
-                quicksink::Action::Close => sender.close().await?
+                quicksink::Action::Close => sender.close().await?,
             }
             Ok(sender)
         });
         let stream = stream::unfold((Vec::new(), receiver), |(mut data, mut receiver)| async {
             match receiver.receive(&mut data).await {
-                Ok(soketto::Incoming::Data(soketto::Data::Text(_))) => {
-                    Some((Ok(IncomingData::Text(mem::take(&mut data))), (data, receiver)))
-                }
-                Ok(soketto::Incoming::Data(soketto::Data::Binary(_))) => {
-                    Some((Ok(IncomingData::Binary(mem::take(&mut data))), (data, receiver)))
-                }
+                Ok(soketto::Incoming::Data(soketto::Data::Text(_))) => Some((
+                    Ok(Incoming::Data(Data::Text(mem::take(&mut data)))),
+                    (data, receiver),
+                )),
+                Ok(soketto::Incoming::Data(soketto::Data::Binary(_))) => Some((
+                    Ok(Incoming::Data(Data::Binary(mem::take(&mut data)))),
+                    (data, receiver),
+                )),
                 Ok(soketto::Incoming::Pong(pong)) => {
-                    Some((Ok(IncomingData::Pong(Vec::from(pong))), (data, receiver)))
+                    Some((Ok(Incoming::Pong(Vec::from(pong))), (data, receiver)))
+                }
+                Ok(soketto::Incoming::Closed(reason)) => {
+                    Some((Ok(Incoming::Closed(reason)), (data, receiver)))
                 }
                 Err(connection::Error::Closed) => None,
-                Err(e) => Some((Err(e), (data, receiver)))
+                Err(e) => Some((Err(e), (data, receiver))),
             }
         });
         Connection {
             receiver: stream.boxed(),
             sender: Box::pin(sink),
-            _marker: std::marker::PhantomData
+            _marker: std::marker::PhantomData,
         }
     }
 
@@ -580,22 +629,20 @@ where
 
 impl<T> Stream for Connection<T>
 where
-    T: AsyncRead + AsyncWrite + Send + Unpin + 'static
+    T: AsyncRead + AsyncWrite + Send + Unpin + 'static,
 {
-    type Item = io::Result<IncomingData>;
+    type Item = io::Result<Incoming>;
 
     fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
         let item = ready!(self.receiver.poll_next_unpin(cx));
-        let item = item.map(|result| {
-            result.map_err(|e| io::Error::new(io::ErrorKind::Other, e))
-        });
+        let item = item.map(|result| result.map_err(|e| io::Error::new(io::ErrorKind::Other, e)));
         Poll::Ready(item)
     }
 }
 
 impl<T> Sink<OutgoingData> for Connection<T>
 where
-    T: AsyncRead + AsyncWrite + Send + Unpin + 'static
+    T: AsyncRead + AsyncWrite + Send + Unpin + 'static,
 {
     type Error = io::Error;
 
